#!/usr/bin/env python3.2
#
# Taglut : Topological And Geometrical Library - a Useful Toolkit
# Copyright (C) 2013 Jean-Marie Favreau <J-Marie.Favreau@udamail.fr>
#                    ISIT, UMR 6284 UdA – CNRS
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License 2
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


"""Usage: meshVeinsPipeline [options] INPUT

Options:
    -x, --x=X                       X coordinate of the center of ROI
    -y, --y=Y                       Y coordinate of the center of ROI
    -z, --z=Z                       Z coordinate of the center of ROI
    -r, --radius=R                  Radius of the ROI
    -i, --output-image=FILE         Save the unfolded mesh as an image
    --sizex=X                       Size of the image (X)
    --sizex=Y                       Size of the image (Y)
    -m, --output-uv=FILE            Save the unfolded mesh as an UV file (format PLY,
                                    containing also the scalar values)
    -h, --help                      Print this message and exit.

INPUT is a mesh file with associated scalar values (available formats: PLY)"""

import taglut
import getopt
import sys

try:
    my_getopt = getopt.gnu_getopt
except AttributeError:
    my_getopt = getopt.getopt


def usage():
    print(__doc__)
    sys.exit(0)

def main():
    try:
        opts, args = my_getopt(sys.argv[1:], "x:y:z:r:i:m:h", ["x=", "y=", "z=",
                                                               "radius=", "output-image=", "output-uv=", "help",
                                                               "sizex=", "sizey="])
    except getopt.GetoptError as msg:
        # print help information and exit:
        print("Error: " + msg)
        usage()
        sys.exit(2)

    sizex=512
    sizey=512
    outputImage = ""
    outputMapping = ""
    x = 0
    y = 0
    z = 0
    r = 0
    coords = 0
    roi = False
    for o, a in opts:
        if o in ("-x", "--x"):
            x = float(a)
            coords += 1
        if o in ("-y", "--y"):
            y = float(a)
            coords += 2
        if o in ("-z", "--z"):
            z = float(a)
            coords += 4
        if o == "sizex":
            sizex = int(a)
        if o == "sizey":
            sizey = int(a)
        if o in ("-r", "--radius"):
            r = float(a)
        if o in ("-i", "--output-image"):
            outputImage = a
        if o in ("-m", "--output-mapping"):
            outputMapping = a
        if o in ("-h", "--help"):
            usage()
            sys.exit()

    try:
        inputFile = args[0]
    except IndexError:
        print("Error: Input file not defined")
        usage()
        sys.exit()

    if coords != 7 or r == 0:
        print("Warning: no RIO defined. Use the full mesh")
    else:
        roi = True

    try:

        # load the mesh
        print("Load mesh and MeshMap")
        plyLoader = taglut.PLYLoader(inputFile)
        mesh = taglut.Mesh()
        mmap = taglut.MeshMap(mesh)
        plyLoader.loadMeshFromPLY(mesh)
        print(" Number of vertices: " + str(mesh.getNbPoints()))
        print(" Number of faces: " + str(mesh.getNbTriangles()))
        plyLoader.loadMeshMapFromPLY(mmap)
        print(" Number of scalar values: " + str(mmap.getNbValues()))

        print("Crop the mesh")
        center = taglut.Point3D(x, y, z)
        mpart = taglut.MeshPart(mesh, center, r, True)

        disc = mpart.buildCropMesh()
        discmap = taglut.MeshMap(disc, mpart, mmap)
        print(" Number of vertices: " + str(disc.getNbPoints()))
        print(" Number of faces: " + str(disc.getNbTriangles()))

        print("Unfold mesh")
        mapping = taglut.Mapping2D3D(disc)
        uMethod = taglut.ABFMethod(mapping)
        mSUnfolding = taglut.MultiscaleUnfolding(False, 1e-8, 30, True)
        nbIter = mSUnfolding.processUnfolding(uMethod, 0)

        print("Save unfolded mesh (" + outputMapping +")")
        plySaver = taglut.PLYSaver()
        plySaver.savePLY(mapping, discmap, outputMapping)

        print("Save unfolded surface as image (" + outputImage +")")
        result = taglut.UCharCImg(sizey, sizey, 1, 3)
        result.fill(0)
        mapping.toImageFromMap(result, discmap)
        result.save(outputImage)

        if nbIter < 0:
            print("Abording.")
            sys.exit(2)

    except taglut.Exception as msg:
        print("Error: " + msg.getMessage())
        sys.exit()


if __name__ == "__main__":
    main()
